import pytest
from swdl import swrest
import json
import os
from swdl.matches import NoPlaylistError, NoEndlistError
from swdl.matches import FFMPegRunner

@pytest.fixture
def data_service():
    cred_file= os.environ.get("SW_CREDENTIAL_FILE","")
    credentials = json.loads(open(cred_file).read())
    account = credentials["username"]
    password = credentials["password"]
    return swrest.DataService(account, password)


def test_get_match(data_service):
    m = data_service.get_match(1300)
    assert m.match_id == 1300


def test_get_match_from_string(data_service):
    m = data_service.get_match("1300")
    assert m.match_id == 1300


def test_update_match(data_service):
    m = data_service.get_match(1300)
    m.pull_info()
    assert m.match_id == 1300


def test_get_matches(data_service):
    m = data_service.get_matches()
    assert len(m) > 0


def test_update_labels(data_service):
    m = data_service.get_match(1300)
    m.pull_labels()
    num_pos = m.labels.positions.shape[0]
    num_events = m.labels.events.shape[0]
    num_status = m.labels.status.shape[0]
    assert num_pos > 1
    assert num_events > 1
    assert num_status > 1
    m.pull_labels()
    assert num_pos == m.labels.positions.shape[0]
    assert num_events == m.labels.events.shape[0]
    assert num_status == m.labels.status.shape[0]

#@pytest.mark.skip(reason="Will take much to long")
def test_download_full_match(data_service):
    m = data_service.get_match(1300)
    m.download_user_stream()

def test_download_match_no_playlist(data_service):
    with pytest.raises(NoPlaylistError):
        m = data_service.get_match(4636)
        m.download_user_stream()

def test_download_match_no_endlist(data_service):
    with pytest.raises(NoEndlistError):
        m = data_service.get_match(4634)
        m.download_grid_stream()

def test_match_endlist(data_service):
    m = data_service.get_match(4634)
    assert m.grid_stream_has_endlist() == False
    assert m.user_stream_has_endlist() == True

    m = data_service.get_match(1300)
    assert m.grid_stream_has_endlist() == True
    assert m.user_stream_has_endlist() == True

def test_find_duration():
    test_string =  " Duration: 02:29:56.61, start: 3630.021300, bitrate: 0 kb/s"
    duration = int(FFMPegRunner.find_duration(test_string))
    assert duration == 2*60*60+29*60+56

def test_find_position():
    test_string="frame=   50 fps=0.0 q=-1.0 Lsize=     776kB time=02:02:02.71 bitrate=2345.6kbits/s speed=2.75x "
    position = int(FFMPegRunner.find_position(test_string))
    assert position == 2*60*60+2*60+2



from tornado.httpclient import HTTPClient, HTTPRequest, HTTPError
from tornado.escape import url_escape
from hashlib import md5
from datetime import datetime
try:
    import json
    from json.decoder import JSONDecodeError
except ImportError:
    import simplejson as json
    from simplejson import JSONDecodeError

from swdl.matches import Match
from swdl.labels import Label
import numpy as np


def to_md5(password):
    """
    Hashes a password to md5

    # Arguments
    password(str): The password to be hashed

    # Returns
    str: Password MD5-hash
    """
    return md5(password.encode()).hexdigest()


class DataService:
    """
    Helper service to download soccerwatch data

    # Attributes
    username (str): Name of the user to connect to the service
    password_hashed (str): MD5 hashed password required to login. Use #to_md5()
    """
    def __init__(self, username, password_hashed):
        self.username=username
        self.user_escaped = url_escape(username)
        self.password = password_hashed
        self.api_stream_url = "https://api-stream-interface-soccerwatch.azurewebsites.net/intern/rest/v1/streamInterface"

        self.client = HTTPClient()

    def get(self, url):
        """
        Performs a get request on the given URL. See alse #DataService.post()

        # Arguments
        url (str): URL to perform get request

        # Retrurns
        str: Dictionary created from JSON dump of the response

        # Example
        ```python
        ds = DataService()
        ds.get("www.google.de")
        ```
        """
        request = HTTPRequest(url, "GET")
        try:
            ret = self.client.fetch(request).body
            return json.loads(ret)
        except JSONDecodeError:
            print("could not parse result")
        except HTTPError as e:
            print("Failed to connect to {}".format(url))

    def post(self, url, **kwargs):
        """
        Performs a post request on the given URL

        # Arguments
        url (str): URL to perform get request
        kwargs: Parameters that will be dumped as JSON

        #Return
        dict: Dictionary created from JSON dump of the response
        """
        header = {"Content-Type": "application/json"}
        request = HTTPRequest(url, "POST", header, json.dumps(kwargs))
        try:
            ret = self.client.fetch(request).body
            return json.loads(ret)
        except JSONDecodeError:
            print("could not parse result")
        except HTTPError as e:
            print("Failed to connect to {}".format(url))

    def get_matches(self):
        """
        Lists all matches

        # Returns
        list: All matches in type #Match`
        """
        url = "{}/match/all/{}/{}".format(self.api_stream_url, self.user_escaped,self.password)
        match_list =  [Match(**x).set_data_service(self) for x in self.get(url)]
        return match_list

    def get_match(self, match_id):
        """
        Returns a single #Match for a given match id

        # Arguments
        match_id (int,str): Match id

        #Returns
        Match: The requested match
        """
        url = "{}/match/single/{}/{}/{}".format(self.api_stream_url, match_id, self.user_escaped, self.password)
        response = self.get(url)
        ret = None if not response else response[0]
        return Match(**ret).set_data_service(self)

    def get_events(self, match_id):
        """
        Returns all Events from azure

        # Arguments
        match_id (str,int): Matchid

        # Returns
        list: Events as got as dictionaries
        """
        url = "{}/tags/{}/{}/{}".format(self.api_stream_url, match_id, self.user_escaped, self.password)
        return self.get(url)

    def get_positions(self, match_id, time=None, virtual_camera_id="-1", source="human"):
        """
        Get the camera positions of a match

        # Arguments
        match_id (str,int): Match id
        time (datetime): Datetime to limit the data given back. Will only return data later than time
        virtual_camera_id (str): The camera id of the positions
        source (str): Should be human or machine

        # Returns
        list: Dictionaries of the positions

        # Example
        """
        if not time:
            last_modified = 0
        elif type(time) == int:
            last_modified = time
        elif type(time) == datetime:
            last_modified = int(time.strftime("%s")) * 1000
        else:
            raise TypeError

        virtual_camera_id = str(virtual_camera_id)
        source = str (source)

        url = "{}/positions/request/{}/{}/{}".format(self.api_stream_url, match_id, self.user_escaped, self.password)
        return self.post(url, virtualCameraId=virtual_camera_id, lastModified=last_modified, source=source)

    def pull_info(self, match):
        """
        Updates the information about a match

        # Arguments
        match (Match): A match

        # Returns
        Match: Updated match

        # Raises
        ValueError: If input is not a match
        """
        if not isinstance(match, Match):
            raise ValueError("Argument must be a valid match")
        return self.get_match(match.match_id)

    def push_labels(self, match, start_index=0, virtual_camera_id="-1", source="human"):
        """
        Uploads the positions greater than the given #start_index to the cloud service

        # Arguments
        match (str): The match
        start_index (int): only positions greater than start index will be pushed
        virtual_camera_id (str): Camera id the positions belongs to
        source (str): Should be "human" or "machine"
        """

        message_body= self._create_label_body(match.match_id, match.positions, start_index, virtual_camera_id, source)
        url = "{}/positions/insert/{}/{}/{}".format(self.api_stream_url,
                                                    match.match_id,
                                                    self.user_escaped,
                                                    self.password)
        self.post(url,**message_body)

    @staticmethod
    def _create_label_body(match_id,  position_list, start_index=0, virtual_camera_id="-1" ,source="human"):

        #ToDo push events and status
        message_body =  dict()
        message_body["virtualCameraId"] = str(virtual_camera_id)
        message_body["matchId"] = str(match_id)
        message_body["source"] = str(source)
        message_body["positions"] = list()
        for i in range(start_index, len(position_list)):
            message_body["positions"].append(dict())
            message_body["positions"][-1]["timestamp"]= str(position_list.positions[0])
            message_body["positions"][-1]["x"]= str(position_list.positions[1])
            message_body["positions"][-1]["y"]= str(position_list.positions[2])
            message_body["positions"][-1]["zoom"]= str(position_list.positions[3])
            message_body["positions"][-1]["source"]= str(source)
        return message_body

    def pull_labels(self, match, time_from=None):
        """
        Updates the labels of a match. Will clear all locally created labels.

        # Arguments
        match (Match): A match
        time_from (datetime,int): Will only get labels later than this time, Must be msecs after 1970 or datetime

        #Returns
        Match: Updated match

        # Raises
        ValueError: If input is not a match
        """
        if not isinstance(match, Match):
            raise ValueError("Argument must be a valid match")

        match_id = match.match_id

        label = Label()
        # Get events
        label.events = np.zeros((0,3), dtype=np.float32)
        label.status = np.zeros((0,2), dtype=np.float32)
        events = self.get_events(match_id)
        for i, e in enumerate(events):
            e_mapped = self.map_events_from_azure(e)
            if len(e_mapped) == 2:
                label.status = np.append(label.status, [e_mapped], axis=0).astype(np.uint32)
            else:
                label.events = np.append(label.events, [e_mapped], axis=0).astype(np.uint32)

        # Get positions
        positions = sorted(self.get_positions(match_id, time=time_from),
                           key=lambda k: k['timestamp'])
        highest_timestamp = int(positions[-1]["timestamp"]) + 1
        label.positions = np.zeros((highest_timestamp, 7), dtype=np.float32)
        for p in positions:
            timestamp = (p["timestamp"])
            x = float(p["x"])
            y = float(p["y"])
            zoom = float(p["zoom"])
            temp = [timestamp * 100, x, y, zoom, x, y, zoom]
            label.positions[timestamp] = temp
        match.labels = label
        return match

    @staticmethod
    def map_events_from_azure(event):
        """
        Maps a event dictionary to a list

        # Arguments
        event (dict): Event dictionary

        # Returns
        list: With either 2 or 3 entries
        """
        e = int(event["eventType"])
        timestamp = int(event["timestamp"])
        if e == 0:
            return [timestamp, 0, 0]
        if e == 1:
            return [timestamp, 0, 1]
        if e == 2:
            return [timestamp, 6, 0]
        if e == 3:
            return [timestamp, 6, 1]
        if e == 4:
            return [timestamp, 5, 0]
        if e == 5:
            return [timestamp, 5, 1]
        if e == 6:
            return [timestamp, 4, 0]
        if e == 7:
            return [timestamp, 4, 1]
        if e == 35:
            return [timestamp, 2, 0]
        if e == 36:
            return [timestamp, 2, 1]
        # For status
        if e == 12:
            return [timestamp * 1000, 1]
        if e == 13:
            return [timestamp * 1000, 2]
        if e == 14:
            return [timestamp * 1000, 3]
        if e == 15:
            return [timestamp * 1000, 4]
        return [timestamp, 7, 0]



import subprocess
from swdl.labels import Label
import m3u8
import urllib
import re

class FFMPegRunner(object):
    """
    Usage:
        runner = FFMpegRunner()
        def status_handler(old, new):
            print "From {0} to {1}".format(old, new)
        runner.run('ffmpeg -i ...', status_handler=status_handler)
    """

    def run(self, command, status_handler=None):
        pipe = subprocess.Popen(command,
            shell=True,
            stdout=subprocess.PIPE,
            stderr=subprocess.STDOUT,
            universal_newlines = True)

        duration = None
        position = None
        percents = 0
        new_pos = 0

        while True:
            line = pipe.stdout.readline().strip()

            if line == '' and pipe.poll() is not None:
                print("I am here")
                break

            if duration is None:
                duration = self.find_duration(line)

            if duration:
                new_pos = self.find_position(line)
                if new_pos:
                    if new_pos != position:
                        position = new_pos
                        if callable(status_handler):
                            status_handler(position, duration)

    @staticmethod
    def find_duration(line):
        re_duration = re.compile('.*Duration: (\d{2}):(\d{2}):(\d{2}).(\d{2})[^\d]*', re.U)
        duration_match=re_duration.match(line)
        if duration_match:
            return FFMPegRunner.time2sec(duration_match)

    @staticmethod
    def find_position(line):
        re_position = re.compile('.*time=(\d{2}):(\d{2}):(\d{2})\.(\d{2})\d*', re.U | re.I)
        position_match = re_position.match(line)
        if position_match:
            return FFMPegRunner.time2sec(position_match)

    @staticmethod
    def time2sec(search):
        secs = sum([int(search.group(i+1))*60**(2-i) for i in range(3)])

        return secs

class SWLoginError(Exception):
    pass


class NoEndlistError(Exception):
    pass


class NoPlaylistError(Exception):
    pass


class Match:
    """
    Class to store match information

    # Attributes
    matchId(str,int): Match id
    swcsId(str,int): The camera id
    name(str): The name of the match. Mostly "TeamA-TeamB"
    location(str): The location where the match took place
    userStreamLink(str): A download link for the user stream
    gridStreamLink(str):A download link for the grid stream
    state(str): Status of the stream. Can be created, live or done

    # Members
    labels(Label): #Label object that stores the labels
    """
    def __init__(self, matchId=None, swcsId="", name="", location="", userStreamLink="", gridStreamLink="", state=""):
        self.match_id=int(matchId)
        self.camera_id=swcsId
        self.name = name
        self.location = location
        self.state = state
        self.user_stream_link = userStreamLink
        self.grid_stream_link = gridStreamLink
        self.labels = Label()
        self.data_service=None

    def copy_info(self, match):
        """
        Copys the info from another match

        # Arguments
        match(Match):
        """
        self.match_id = match.match_id
        self.camera_id = match.camera_id
        self.name = match.name
        self.location = match.location
        self.state = match.state
        self.user_stream_link = match.user_stream_link
        self.grid_stream_link = match.grid_stream_link

    def download_user_stream(self, download_directory="."):
        """
        Downloads the user stream from the match. Needs ffmpeg

        # Attributes
        download_directory(str):

        # Raises
        NoPlaylistError: If video file is not existing
        NoEndlistError: If stream has not endlist tag. This means the stream is not correctly finished and can note be downlaoded with ffmpeg
        """
        self._download_match(download_directory, "User")

    def download_grid_stream(self, download_directory="."):
        """
        Downloads the grid stream from the match. Needs ffmpeg

        # Attributes
        download_directory:

        # Raises
        NoPlaylistError: If video file is not existing
        NoEndlistError: If stream has not endlist tag. This means the stream is not correctly finished and can note be downlaoded with ffmpeg
        """
        self._download_match(download_directory, "Grid")

    def pull_info(self):
        """
        Updates match info

        # Raises
        SWLoginError
        """
        if not self.data_service:
            raise SWLoginError("Not logged into Soccerwatch Data Service")
        m = self.data_service.pull_info(self)
        self.copy_info(m)

    def pull_labels(self):
        """
        Updates match labels.
        Warning: It will override all lokally stored labels

        # Raises
        SWLoginError: If data service was not connected
        """
        if not self.data_service:
            raise SWLoginError("Not logged into Soccerwatch Data Service")
        m = self.data_service.pull_labels(self)
        self.labels = m.labels

    def push_labels(self):
        """
        Uploads the labels to the cloud

        # Raises
        SWLoginError: If data service was not connected
        """
        if not self.data_service:
            raise SWLoginError("Not logged into Soccerwatch Data Service")
        self.data_service.push_labels(self)

    def set_data_service(self, data_service):
        """
        Sets the #DataService for this match

        # Arguments
        data_service(DataService):
        """
        self.data_service = data_service
        return self

    def grid_stream_available(self):
        """
        Checks if a user stream is available

        # Returns
        bool: if available
        """
        return self._check_stream("User")

    def user_stream_available(self):
        """
        Checks if a user stream is available

        # Returns
        bool: if available
        """
        return self._check_stream("Grid")

    def user_stream_has_endlist(self):
        """
        Checks if user stream has an an endlist tag, which means that the stream has been completely uploaded

        # Returns

        bool: True if endlist is exsisting
        """
        return self._check_endlist("User")

    def grid_stream_has_endlist(self):
        """
        Checks if grid stream has an an endlist tag, which means that the stream has been completely uploaded

        # Returns

        bool: True if endlist is exsisting
        """
        return self._check_endlist("Grid")

    def _check_stream(self, streamType="User"):
        if streamType == "User":
            download_link = self.user_stream_link
        else:
            download_link = self.grid_stream_link
        try:
            playlist = m3u8.load(download_link)
        except urllib.error.HTTPError:
            return False
        return True

    def _check_endlist(self, streamType="User"):
        if streamType == "User":
            download_link = self.user_stream_link
        else:
            download_link = self.grid_stream_link
        try:
            playlist = m3u8.load(download_link)
            if not playlist.is_endlist:
                return False
        except urllib.error.HTTPError:
            return False
        return True


    def _download_match(self, download_directory=".", streamType="User"):
        hint=""
        if streamType == "User":
            hint="user"
            download_link = self.user_stream_link
        else:
            hint="grid"
            download_link = self.grid_stream_link
        # Check m3u8
        try:
           playlist = m3u8.load(download_link)
           if not playlist.is_endlist:
               raise NoEndlistError("Video has no endlist, can not be downloaded with FFMPEG")
        except urllib.error.HTTPError:
               raise NoPlaylistError("Playlist was not found")
        command = "ffmpeg -i {} -c copy -y \"{}/{}_{}_{}.mp4\"".format(download_link,
                                                                   download_directory,
                                                                   self.match_id,
                                                                   self.name,
                                                                   hint)
        print(command)
        def run_callback(actual,total):
            percent = 100.0*actual/total
            print("%.2f%%"%percent)

        ffmpeg = FFMPegRunner()
        ffmpeg.run(command, status_handler=run_callback)


    def __str__(self):
        return "Match[\tid:\t\t{0},\n\tcamera:\t\t{1},\n\tname:\t\t{2},\n\tlocation:\t{3}" \
               "\n\tstate:\t\t{4}\n\tvideo:\t\t{5}\n\tgrid:\t\t{6}]".format(
            self.match_id,
            self.camera_id,
            self.name,
            self.location,
            self.state,
            self.user_stream_link,
            self.grid_stream_link)



import h5py as h5
import numpy as np

class Label:
    """
    Structure to store label data, just wraping numpy arrays
    """
    def __init__(self):
        self.events = np.zeros((0,3),dtype=np.uint32)
        self.status = np.zeros((0,2),dtype=np.uint32)
        self.positions = np.zeros((0,7),dtype=np.float32)

    @classmethod
    def from_file(cls, path="labels.h5"):
        """
        Reads from hdf5 file

        # Attributes
        path(str):
        """
        file = h5.File(path,"r")
        label = cls()
        label.positions = file["labels"][:]
        label.events = file["events"][:]
        label.status = file["status"][:]
        file.close()
        return label

    def save(self,path="labels.h5"):
        """
        Saves label to hdf5

        # Attributes
        path(str):
        """
        file = h5.File(path,"w")
        file["events"] = self.events
        file["labels"] = self.positions
        file["status"] = self.status
        file.close()


from setuptools import setup

setup(name='swdl',
      version='0.1',
      description='Soccerwatch Data Library',
      author='Christian Bodenstein',
      author_email='bodenstein@soccerwatch.tv',
      packages=['swdl'],      
      install_requires=[
          'numpy',
          'h5py',
          'tornado',
          'm3u8',
          'pydoc-markdown'
      ],
      zip_safe=False)
